#include "combine.h"
#include <stdio.h>
#include "mpi.h"
#include <stdlib.h>

/*
 * All-Gather
 * Code starts with n Elements and finishes with n*p (N) elements.
 * The result is an array containing all elements 
 */

// Ring All-Gather
int ring_gather(double *sendbuf, double *recvbuf, 
struct args *args, int rank, int size) {
  MPI_Request request;
  MPI_Status status;
  double *tmpbuf;
  int i, j;
  int curpos;
  int tag=0;

  tmpbuf = (double*)malloc(args->numelem*sizeof(double));
  curpos = rank;

  // sendbuf is the data of stuff which I already have.
  // tmpbuf contains the data to be passed on. recvbuf contains 
  // the data containing the result array.

  for (i=0;i<args->numelem;i++) {
    // first data to pass on is my own sendbuf data.
    tmpbuf[i] = sendbuf[i];
    // of course I need to allocate my own sendbuf results to recvbuf
    recvbuf[curpos*(args->numelem)+i] = tmpbuf[i];
  }
  
  // curpos keeps track of which portion of I am receiving at the moment

  if (curpos == 0) {
    curpos = size-1;
  } else {
    curpos--;
  }

  for(i=1;i<size;i++) {
    tag = i;
    // Everybody sends the current data that they have
    MPI_Isend(tmpbuf, args->numelem, MPI_DOUBLE, (rank+1)%size, tag, 
              MPI_COMM_WORLD, &request);
    if (!rank) {
      // Everybody receives data from the previous rank
      MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, size-1, tag, 
                MPI_COMM_WORLD, &request);
    } else {
      MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, rank-1, tag, 
                MPI_COMM_WORLD, &request);
    }
    // Wait to get all data
    MPI_Wait(&request, &status);
    
    // Stick data into recvbuf using curpos to keep track
    for(j=0;j<args->numelem;j++) {
      recvbuf[curpos*(args->numelem)+j] = tmpbuf[j];
    }

    // Do computation 
    if (curpos == 0) {
      curpos = size-1;
    } else {
      curpos--;
    }
  }

  // Some stuff to print out (based on -o argument)

  if((!rank) && (args->output)){
    for (i=0;i<args->numelem*size;i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  }
  free(tmpbuf);
  return 0;
}

// Fan-in Fan-out All-Gather
int fan_gather(double *sendbuf, double *recvbuf, 
struct args *args, int rank, int size) {
  MPI_Request request;
  MPI_Status status;
  int i; 
  int tag=0;
  int finished;
  int offset;
  int checkrank;
  int sent = FALSE;
  double *tmpbuf;

  tmpbuf = (double*)malloc(size*args->numelem*sizeof(double));
  offset = 1;

  for (i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }
  
  if (size > 1) {
    // initialise tree traversing variables
    finished = FALSE;
    offset = 1;
    checkrank = rank;

    // Fan-in

    while(!finished) {
      // If this node is to receive
      if (checkrank%2 == 0) {
	// receive from defined offset
        if (rank+offset < size) {
          if (!sent) {
            MPI_Irecv(tmpbuf, 
                      args->numelem*offset, MPI_DOUBLE, rank+offset, 
                      tag, MPI_COMM_WORLD, &request);
            MPI_Wait(&request, &status);
	    // Save received data
            for (i=0;i<args->numelem*offset;i++) {
              recvbuf[offset*(args->numelem)+i] = tmpbuf[i];
            }
          }
        }
      } else {
	// if my checkrank is odd numbered, I will send my data
        if (!sent) {
          MPI_Isend(recvbuf, args->numelem*offset,  
                    MPI_DOUBLE, rank-offset, tag, MPI_COMM_WORLD, &request);
          MPI_Wait(&request, &status);
          sent = TRUE;
        } 
        // After sending my data, I finish
      }
   
      tag++;
      checkrank = checkrank/2;
      offset = offset*2;
      
      if (offset >= size) finished = TRUE;
      
      
    }
    
  }
  
  // Fan-out
  // By this time the root node will have the combined data.

  finished = FALSE;

  while (!finished) {
    // If I am to send , send my data
    if (rank%offset == 0) {
      if ((rank+offset/2) < size) {
        MPI_Isend(recvbuf, args->numelem*size, MPI_DOUBLE, rank+offset/2, tag, 
                 MPI_COMM_WORLD, &request);
      }
    }
    // If I am to receive, receive data
    if((rank-offset/2)%offset == 0) {
      if ((rank-offset/2) >= 0) {
        MPI_Irecv(recvbuf, args->numelem*size, MPI_DOUBLE, rank-offset/2, tag, 
                 MPI_COMM_WORLD, &request);
      }
    }
    MPI_Wait(&request, &status);
    // Change the offset
    if (offset == 1) {
      finished = TRUE;
    } else {
      offset = offset/2;
    }
  }

  // If need be, print out the results

  if((!rank) && (args->output)){
    for (i=0;i<args->numelem*size;i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  }
  free(tmpbuf);
  return 0;
}


// Bi-Directional All-Gather
int bidir_gather(double *sendbuf, double *recvbuf, struct args *args, 
                 int rank, int size) {

  int offset;
  int finished = FALSE;
  int send;
  int sender = FALSE;
  int i;
  MPI_Status status;
  MPI_Request request;
  int tag=0;
  double *tmpbuf;
  double *tmpbuf2;
  int curpos;
  int j;

  send = 0;
  offset = 1;
  curpos = rank;
  
  tmpbuf = (double*)malloc(args->numelem*size*sizeof(double));
  tmpbuf2 = (double*)malloc(args->numelem*size*sizeof(double));

  for(i=0;i<args->numelem;i++) recvbuf[rank*args->numelem+i] = sendbuf[i];

  finished = FALSE;

  // Similar to Fan-in Fan-out except there is no fan-out stage
  while(!finished) {
    // A sender sends to a non-sender and vice versa. A sender does not send to
    // another sender. The term sender is used to differentiate pairs of nodes.
    // offset determines the difference in node rank to send to. 1, 2, 4 etc.
    for (i=0;i<size;i++) {
      if (!(i % offset)) {
        send ^= 1;
      }
      if ((rank == i) && send) {
        sender = TRUE;
      }
    }

    for (i=0;i<args->numelem*offset;i++) {
      tmpbuf2[i] = recvbuf[curpos*args->numelem+i];
    }

    // Do the data exchange
    if(sender) {
      MPI_Isend(tmpbuf2, args->numelem*offset, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &request);
      MPI_Irecv(tmpbuf, args->numelem*offset, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &request);
      MPI_Wait(&request, &status);

      for (i=0;i<args->numelem*offset;i++) {
        recvbuf[(curpos+offset)*(args->numelem)+i] = tmpbuf[i];
      }    
    } else {
      MPI_Isend(tmpbuf2, args->numelem*offset, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD,&request);
      MPI_Irecv(tmpbuf, args->numelem*offset, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD, &request);
      MPI_Wait(&request, &status);
      for (i=0;i<args->numelem*offset;i++) {
        recvbuf[(curpos-offset)*(args->numelem)+i] = tmpbuf[i];
      }
      curpos = curpos-offset;
    }
    
    tag++;

    // change offset and continue until there is no more data to send
    offset *= 2;
    if(offset >= size) {
      finished = TRUE;
    } else {
      sender = FALSE;
    }
  } 
  if((rank==args->testrank) && (args->output)){
    for (j=0;j<size;j++) {
      for (i=0;i<args->numelem;i++) {
        printf(" %f ", recvbuf[j*args->numelem+i]);
      }
      printf("\n");
    }
  }

  free(tmpbuf);
  free(tmpbuf2);
  return 0;


}

// Recursive-Halving Recursive-Doubling All-Gather (extremely complicated code)
// Will try to describe in as much detail as possible.
int rec_gather(double *sendbuf, double *recvbuf, struct args *args, 
               int rank, int size) {
 
  int *startelem;
  int done = FALSE;
  int offset;
  int send = 0;
  int sender = FALSE;
  int leftover;
  int *blocksize;
  int tmpblock;
  int i,j,k;
  int sendsize;
  double *tmpbuf;
  double *tmpbuf2;
  int tag = 0;
  int activepos;
  int notactivepos;
  int curpos;
  int numsend;
  int numrecv;
  int count;
  MPI_Status status;
  MPI_Request request;

  curpos = rank;
 
  leftover = args->numelem % size;
  tmpblock = args->numelem / size;

  blocksize = (int*)malloc(size*sizeof(int));
  startelem = (int*)malloc(size*sizeof(int));
  tmpbuf = (double*)malloc(args->numelem*size*sizeof(double));
  tmpbuf2 = (double*)malloc(args->numelem*size*sizeof(double));

  // I first determine the blocksize. A block is numelem / size. 

  for (i=0;i<size;i++) {
    blocksize[i] = tmpblock;
    if (leftover > 0) {
      blocksize[i]++;
      leftover--;
    }
  }
  
  // This part determines the start of each block. This makes it easy to access
  // the block later.
  startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + blocksize[i-1];
  }

  // I initialise receive elements to send elements.
  for(i=0;i<args->numelem;i++) {
    recvbuf[rank*args->numelem+i] = sendbuf[i];
  }

  // Recursive-Halving

  // activepos points to the start of the blcoks I am going to send. 
  // I assume the activepos for all nodes to be 0 first.
  activepos = 0;
  sendsize = size/2;
  send = 0;
  offset = 1;

  while (!done) {
    // Like Bi-Directional exchange, I will pair up the nodes as 
    // sender or non-sender. Both sender and non-sender nodes send, it 
    // is just a definition.
    for (i=0;i<size;i++) {
      if (!(i % offset)) {
        send ^= 1;
      }
      if ((rank == i) && send) {
        sender = TRUE;
      }
    }

    // Sender nodes send the top half of their data (activepos=0). 
    // Non-sender nodes send the bottom half of their data (activepos=size/2)
    // notactivepos is the position to receive the data from.
    // notactivepos is the opposite of activepos. And is the value of old 
    // activepos
    if (sender) {
      notactivepos = activepos+sendsize;
    } else {
      notactivepos = activepos;
      activepos = activepos + sendsize;
    }

    numsend = 0;
    numrecv = 0;
    count = 0;

    // This is the tricky bit. I need to know exactly what information to send.
    // Therefore I calculate the offset to know the node this local node is 
    // sending to. Then I calculate the number of blocks to send (sendsize) and
    // finally save all this information in tmpbuf while also counting the 
    // number of elements I have saved (count).
    for (j=0;j<offset;j++) {
      for (i=0;i<sendsize;i++) {
        numsend = numsend + blocksize[activepos+i];
        for (k=0;k<blocksize[activepos+i];k++) {
          tmpbuf[count] = 
            recvbuf[(curpos+j)*args->numelem+startelem[activepos+i]+k];
          count++;
        }
      }
    }

    // I also need to know how many elements I will be receiving.
    for (j=0;j<offset;j++) {
      for (i=0;i<sendsize;i++) {
        numrecv = numrecv + blocksize[notactivepos+i];
      }
    }
 
    // Once all that information is prepared, I send it.
    // I also receive the information from the remote node. 
    if (sender) {
      MPI_Isend(tmpbuf, numsend, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD,&request);
      MPI_Irecv(tmpbuf2, numrecv, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &request);
      MPI_Wait(&request, &status);
      count = 0;
      // This part is a reverse of the data preparation. I save the information
      // in the correct memory address in recvbuf. I again use offset, sendsize
      // and blocksize to calculate this address.
      for(j=0;j<offset;j++) {
        for(i=0;i<sendsize;i++) {
          for(k=0;k<blocksize[notactivepos+i];k++) {
            recvbuf[(curpos+offset+j)*args->numelem
                   +startelem[notactivepos+i]
                   +k] = tmpbuf2[count];
            count++;
          }
        }
      }
            
      // Same as previous except this is for the 'non-sender' node
    } else {
      MPI_Isend(tmpbuf, numsend, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD,&request);
      MPI_Irecv(tmpbuf2, numrecv, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD, &request);
      MPI_Wait(&request, &status);
      count = 0;
      for(j=0;j<offset;j++) {
        for(i=0;i<sendsize;i++) {
          for(k=0;k<blocksize[notactivepos+i];k++) {
            recvbuf[(curpos-offset+j)*args->numelem
                   +startelem[notactivepos+i]
                   +k] = tmpbuf2[count];
            count++;
          }
        }
      }
      curpos = curpos - offset;
    }
    // If I am down to one block, I can finish my code.
    if (sendsize == 1) {
      done = TRUE;
    } else {
      // else swap activepos and notactivepos
      // half the sendsize
      // increase offset by 2 times
      // reinitialise sender and send
      activepos = notactivepos;
      sendsize = sendsize / 2;
      offset = offset * 2;
      sender = 0;
      send = 0;    
    }
  }

  // Recursive-Doubling

  // reinitalise data in preparation for recursive-doubling
  // By this time each node has a segment of completed data.
  // This is initially defined by notactivepos which is changed to activepos
  done = FALSE;
  sender = FALSE;
  send = 0;

  activepos = notactivepos;

  // pair up sender non-sender pairs.
  for (i=0;i<size;i++) {
    if (!(i % offset)) {
      send ^= 1;
    }
    if ((rank == i) && send) {
      sender = TRUE;
    }
  }
  
  // calculate new notactivepos. 
  if (sender) {
    notactivepos = activepos - sendsize;
  } else {
    notactivepos = activepos + sendsize;
  }

  while(!done) {
    count = 0;
    numsend = 0;
    // again I initialise data based on the remote node.
    // the difference is now each node has a complete set of data. Which is why
    // the entire size is used.
    for (j=0;j<size;j++) {
      for (i=0;i<sendsize;i++) {
        numsend = numsend + blocksize[activepos+i];
        for (k=0;k<blocksize[activepos+i];k++) {
          tmpbuf[count] =
            recvbuf[j*args->numelem
                   +startelem[activepos+i]
                   +k];
          count++;
        }
      }
    }

    // Calculate the amount of eleemnts I am to receive.
    numrecv = 0;
    for (j=0;j<size;j++) {
      for (i=0;i<sendsize;i++) {
        numrecv = numrecv + blocksize[notactivepos+i];
      }
    }
    
    // Do the data exchange
    if (sender) {
      MPI_Isend(tmpbuf, numsend, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD,&request);
      MPI_Irecv(tmpbuf2, numrecv, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &request);
      MPI_Wait(&request, &status);

      count=0;
      // save the received data in the correct memory address
      for (j=0;j<size;j++) {
        for (i=0;i<sendsize;i++) {
          for (k=0;k<blocksize[notactivepos+i];k++) {
            tmpbuf[count] =
              recvbuf[j*args->numelem
                     +startelem[notactivepos+i]
                     +k] = tmpbuf2[count];
            count++;
          }
        }
      }
    } else {
      MPI_Isend(tmpbuf, numsend, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD,&request);
      MPI_Irecv(tmpbuf2, numrecv, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD, &request);
      MPI_Wait(&request, &status);

      count = 0;
      for (j=0;j<size;j++) {
        for (i=0;i<sendsize;i++) {

          for (k=0;k<blocksize[notactivepos+i];k++) {
            recvbuf[j*args->numelem
                   +startelem[notactivepos+i]
                   +k] = tmpbuf2[count];
            count++;
          }
        }
      }

    }
    
    // recalculate sendsize, sender and activepos values.
    if (sendsize == size/2) {
      done = TRUE;
    } else {
      if (sender) activepos = notactivepos;
        
      sendsize = sendsize * 2;
      offset = offset / 2;
      sender = 0;
      send = 0;
    

      for (i=0;i<size;i++) {
        if (!(i % offset)) {
          send ^= 1;
        }
        if ((rank == i) && send) {
          sender = TRUE;
        }
      }
      
      if (sender) {
        notactivepos = activepos - sendsize;
      } else {
        notactivepos = activepos + sendsize;
      }
    }
  }

  // Print out results if -o argument is defined

  if((rank==args->testrank) && (args->output)){
    for (j=0;j<size;j++) {
      for (i=0;i<args->numelem;i++) {
        printf(" %f ", recvbuf[j*args->numelem+i]);
      }
      printf("\n");
    }
  }

  free(startelem);
  free(blocksize);
  free(tmpbuf);

  return 0;
}

// Bucket All-Gather (OBSOLETE)

int buck_gather(double *sendbuf, double *recvbuf, struct args *args, 
                int rank, int size) {
  int * blocksize;
  int * startelem;
  double * tmpbuf;
  double * tmpbuf2;
  int i, j , k;
  MPI_Status status;
  MPI_Request request;
  int leftover;
  int sendsize;
  int curblock;
  int currecvblock;
  int tmpblock;
  int tag=0;

  blocksize = (int*)malloc(size*sizeof(int));
  startelem = (int*)malloc(size*sizeof(int));
  tmpbuf = (double*)malloc(args->numelem*size*sizeof(double));
  tmpbuf2 = (double*)malloc(args->numelem*size*sizeof(double));

  leftover = args->numelem % size;
  tmpblock = args->numelem / size;

  for (i=0;i<size;i++) {
    blocksize[i] = tmpblock;
    if (leftover > 0) {
      blocksize[i]++;
      leftover--;
    }
  }
  
  startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + blocksize[i-1];
  }

  for(i=0;i<args->numelem;i++) {
    recvbuf[rank*args->numelem+i] = sendbuf[i];
  }

  sendsize = 1;
  curblock = rank;
  currecvblock = (size + (rank-1)) % size;
  
  for(i=0;i<((size-1)*2);i++) {
    for(k=0;k<sendsize;k++) {
      for(j=0;j<blocksize[curblock];j++) 
        tmpbuf2[k*blocksize[curblock]+j] = 
          recvbuf[((size+(rank-k))%size)*args->numelem
                 +startelem[curblock]+j];    
    }
    sd(rank, (rank+1)%size);
    MPI_Isend(tmpbuf2, sendsize*blocksize[curblock], MPI_DOUBLE, 
              (rank+1)%size, 
              tag, MPI_COMM_WORLD,&request);
    sd2(rank, (rank+1)%size);
    if (!rank) {
      rd(rank, size-1);
      MPI_Irecv(tmpbuf, sendsize*blocksize[currecvblock], MPI_DOUBLE, size-1, 
                tag, 
                MPI_COMM_WORLD, &request);
    } else {
      rd(rank, rank-1);
      MPI_Irecv(tmpbuf, sendsize*blocksize[currecvblock], MPI_DOUBLE, rank-1, 
                tag, 
                MPI_COMM_WORLD, &request);
    }
    tag++;

    MPI_Wait(&request, &status);
    rd2(rank, rank-1);

    for(k=0;k<sendsize;k++) {
      for(j=0;j<blocksize[currecvblock];j++) 
        recvbuf[((size+(rank-k-1))%size)*args->numelem
               +startelem[currecvblock]+j] = 
          tmpbuf[k*blocksize[currecvblock]+j];    
    }
      
    
    curblock = currecvblock;
    currecvblock = (size + (currecvblock-1)) % size;
    if (sendsize < (size - 1)) sendsize++;
    
  }

  if((rank==args->testrank) && (args->output)){
    for (j=0;j<size;j++) {
      for (i=0;i<args->numelem;i++) {
        printf(" %f ", recvbuf[j*args->numelem+i]);
      }
      printf("\n");
    }
  }

  return 0;
  
}

// MPI All-Gather Implementation 

int mpi_gather(double *sendbuf, double *recvbuf, struct args *args, 
                int rank, int size) {

  int i, j;

  MPI_Allgather(sendbuf, args->numelem, MPI_DOUBLE, recvbuf, 
                args->numelem, MPI_DOUBLE, MPI_COMM_WORLD);
  
  if((rank==args->testrank) && (args->output)){
    for (j=0;j<size;j++) {
      for (i=0;i<args->numelem;i++) {
        printf(" %f ", recvbuf[j*args->numelem+i]);
      }
      printf("\n");
    }
  }

  return 0;
}

// Pipeline All-Gather, Repeated Binary Tree All-Gather
// Pipeline and Repeated Binary Tree was initally considered as the same
// algorithm with the only difference being the method of broadcast
// This is not correct and given time I would have implemented these two 
// algorithms as separate functions.

int pipe_gather(double *sendbuf, double *recvbuf, struct args *args, 
                int rank, int size) {
  double * tmpbuf;
  int i, j, k;
  MPI_Status status;
  MPI_Request request;
  int tag=0;
  int sender;
  int receiver;
  int offset;
  int count;
  int finished;
  int curnode;

  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for(i=0;i<args->numelem;i++) {
    recvbuf[rank*args->numelem+i] = sendbuf[i];
  } 

  // Do Pipeline All-Gather

  if (args->ring == TRUE) {

    // i in this for loop defines the current active node (node to send)
    for (i=0;i<size;i++) {
      // active node sends
      if (rank == i) {
        MPI_Isend(sendbuf, args->numelem, MPI_DOUBLE, (rank+1)%size,
                  tag, MPI_COMM_WORLD,&request); 
        MPI_Wait(&request, &status);
      } else {
	// everyone else receives the data active node is sending
        MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, (rank-1+size)%size, 
                  tag, MPI_COMM_WORLD, &request); 
        MPI_Wait(&request, &status);  
        for (j=0;j<args->numelem;j++) { 
          recvbuf[i*args->numelem+j] = tmpbuf[j]; 
        }
	// If I still have previously received data, send it
        if (rank != (i+size-1)%size) {
          MPI_Isend(tmpbuf, args->numelem, MPI_DOUBLE, (rank+1)%size,
                    tag, MPI_COMM_WORLD,&request); 
          MPI_Wait(&request, &status);  
        }
      }        
    }
  } else {
    // Do Repeated Binary Tree
    // Again i in the for loop is used to define the active node
    for (i=0;i<size;i++) {
      sender = FALSE;
      receiver = FALSE;
      offset = size / 2;
      // active node is sender (sender, non-sender pairs)
      if (rank == i) {
        sender = TRUE;
      }
      // active + offset node is the receiver
      if (rank == ((i + offset) % size)) {
        receiver = TRUE;
      }
      count = 1;

      finished = FALSE;

      // This is the binary tree part.
      while (!finished) {
	// sender sends, receiver receives
        if (sender) {
          for (k=0;k<args->numelem;k++) {
            tmpbuf[k] = recvbuf[i*args->numelem+k];
          }
          MPI_Isend(tmpbuf, args->numelem, MPI_DOUBLE, 
                    (rank + offset) % size,
                    tag, MPI_COMM_WORLD,&request);
          MPI_Wait(&request, &status); 
        }
        if (receiver) {
          MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, 
                    (rank + size - offset) % size, tag,  
                    MPI_COMM_WORLD, &request); 
          MPI_Wait(&request, &status);
          for (k=0;k<args->numelem;k++) {
            recvbuf[i*args->numelem+k] = tmpbuf[k];
          }
        }
	// recalculate sender and receiver for the same active node

        if (offset == 1) {
          finished = TRUE;
        } else {
          if (receiver) {
            sender = TRUE;
            receiver = FALSE;
          }   
          offset = offset/2;
          count = count * 2;
          curnode = (i + offset) % size;
          for (j=0;j<count;j++) {
            if (rank ==  curnode) {
              receiver = TRUE;
            }
            curnode = (curnode + (offset*2)) % size;
          }
        }
      }
    }    
  }

  // print out the results if -o argument is defined

  if((rank==args->testrank) && (args->output)){
    for (j=0;j<size;j++) {
      for (i=0;i<args->numelem;i++) {
        printf(" %f ", recvbuf[j*args->numelem+i]);
      }
      printf("\n");
    }
  }      
  
  return 0;

}

// Full Fan-In All-Gather (initially called Bing's Special)

int bing_gather(double *sendbuf, double *recvbuf, struct args *args, 
                 int rank, int size) {

  int i,j,k;
  double * tmpbuf;
  int tag=0;
  MPI_Request request;
  MPI_Status status;

  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for (i=0;i<args->numelem;i++) {
    recvbuf[rank*args->numelem+i] = sendbuf[i];
  }
  
  // i defines which node is receiving (active node) and j defines
  // which node is sending

  for (i=0;i<size;i++) {
    for(j=1;j<size;j++) {
      // If I am active, receive from everyone else
      if (rank == i) {

        MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, 
                  (rank + j) % size, tag, 
                  MPI_COMM_WORLD, &request);
        MPI_Wait(&request, &status);
        for (k=0;k<args->numelem;k++) {
          recvbuf[((rank + j) % size)*args->numelem+k] = tmpbuf[k];
        }
      } 
      
      // everyone else sends
      if (rank == ((i + j) % size)) {

        MPI_Isend(sendbuf, args->numelem, MPI_DOUBLE, 
                  (rank-j+size)%size, tag, 
                  MPI_COMM_WORLD, &request);
        MPI_Wait(&request, &status);

      }
    }
  }

  // print out results if -o argument is defined

  if((rank==args->testrank) && (args->output)){
    for (j=0;j<size;j++) {
      for (i=0;i<args->numelem;i++) {
        printf(" %f ", recvbuf[j*args->numelem+i]);
      }
      printf("\n");
    }
  }      
  return 0;

}

// Full Fan-Out All-Gather (Initially called Naive Algorithm)

int naive_gather(double *sendbuf, double *recvbuf, struct args *args, 
		 int rank, int size) {
  int i, j;
  int tag;
  double * tmpbuf;
  MPI_Status status;
  MPI_Request request;

  tag = 0;

  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for(i=0;i<args->numelem;i++) {
    recvbuf[rank*(args->numelem)+i] = sendbuf[i];
  }
 
  // send my data to every other node
  for(i=1;i<size;i++) {
    MPI_Isend(sendbuf, args->numelem, MPI_DOUBLE, (rank + i)%size,
	      tag, MPI_COMM_WORLD, &request);
  }

  // receive my data from every other node
  for(i=1;i<size;i++) {
    MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, (rank + i)%size, 
	      tag, MPI_COMM_WORLD, &request);
    MPI_Wait(&request, &status);
    for (j=0;j<args->numelem;j++) {
      recvbuf[((rank + i) % size)*args->numelem+j] = tmpbuf[j];
    }
  }

  // print out the results f -o argument is defined
  
  if((rank==args->testrank) && (args->output)){
    for (j=0;j<size;j++) {
      for (i=0;i<args->numelem;i++) {
        printf(" %f ", recvbuf[j*args->numelem+i]);
      }
      printf("\n");
    }
  }      
  return 0;
}

// ! NEED WORK HERE !

int fourway_gather(double *sendbuf, double *recvbuf, struct args *args, 
		 int rank, int size) {
  if (!rank) printf("doing 4 way gather\n");

  return 0;
}

//---------------------------------------------
// FROM THIS POINT ONWARDS, THE CODE USES BLOCKING ALGORITHMS 
// AND IS OBSOLETE

int b_ring_gather(double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, 
                  int size) {
  MPI_Status status;
  double *tmpbuf;
  int i, j;
  int curpos;
  int tag=0;

  tmpbuf = (double*)malloc(args->numelem*sizeof(double));
  curpos = rank;

  for (i=0;i<args->numelem;i++) {
    tmpbuf[i] = sendbuf[i];
    recvbuf[curpos*(args->numelem)+i] = tmpbuf[i];
  }
  
  if (curpos == 0) {
    curpos = size-1;
  } else {
    curpos--;
  }

  for(i=1;i<size;i++) {
    tag = i;
    
    if (!rank) {     
      MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, size-1, tag, 
                MPI_COMM_WORLD, &status);
      MPI_Send(tmpbuf, args->numelem, MPI_DOUBLE, (rank+1)%size, tag, 
              MPI_COMM_WORLD);
    } else {
      MPI_Send(tmpbuf, args->numelem, MPI_DOUBLE, (rank+1)%size, tag, 
              MPI_COMM_WORLD);
      MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, rank-1, tag, 
                MPI_COMM_WORLD, &status);
    }
    for(j=0;j<args->numelem;j++) {
      recvbuf[curpos*(args->numelem)+j] = tmpbuf[j];
    }

    // Do computation 
    if (curpos == 0) {
      curpos = size-1;
    } else {
      curpos--;
    }
  }

  if((!rank) && (args->output)){
    for (i=0;i<args->numelem*size;i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  }
  free(tmpbuf);
  return 0;
}


int b_fan_gather(double *sendbuf, double *recvbuf, struct args *args, int rank, 
                      int size) {
  MPI_Status status;
  int i; 
  int tag=0;
  int finished;
  int offset;
  int checkrank;
  int sent = FALSE;
  double *tmpbuf;

  tmpbuf = (double*)malloc(size*args->numelem*sizeof(double));
  offset = 1;

  for (i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }
  
  if (size > 1) {
    // initialise tree traversing variables
    finished = FALSE;
    offset = 1;
    checkrank = rank;

    // Fan-in

    //MPI_Barrier(MPI_COMM_WORLD);

    while(!finished) {
      
      if (checkrank%2 == 0) {
        if (rank+offset < size) {
          if (!sent) {
            MPI_Recv(tmpbuf, 
                      args->numelem*offset, MPI_DOUBLE, rank+offset, 
                     tag, MPI_COMM_WORLD, &status);
            for (i=0;i<args->numelem*offset;i++) {
              recvbuf[offset*(args->numelem)+i] = tmpbuf[i];
            }
          }
        }
      } else {
	// if my checkrank is odd numbered, I will send my data
        if (!sent) {
          MPI_Send(recvbuf, args->numelem*offset,  
                    MPI_DOUBLE, rank-offset, tag, MPI_COMM_WORLD);
          sent = TRUE;
        } 
        // After sending my data, I finish
      }
      
      tag++;
      checkrank = checkrank/2;
      offset = offset*2;
      
      if (offset >= size) finished = TRUE;
      
      
    }
    
  }
  
  finished = FALSE;

  while (!finished) {
    if (rank%offset == 0) {
      if ((rank+offset/2) < size) {
        //rank%offset : send(rank+offset/2);
        MPI_Send(recvbuf, args->numelem*size, MPI_DOUBLE, rank+offset/2, tag, 
                 MPI_COMM_WORLD);
      }
    }
    if((rank-offset/2)%offset == 0) {
      if ((rank-offset/2) >= 0) {
        //((rank-offset/2)%offset) : recv(rank-offset/2);
        MPI_Recv(recvbuf, args->numelem*size, MPI_DOUBLE, rank-offset/2, tag, 
                 MPI_COMM_WORLD, &status);
      }
    }
    if (offset == 1) {
      finished = TRUE;
    } else {
      offset = offset/2;
    }
    
  }

  if((!rank) && (args->output)){
    for (i=0;i<args->numelem*size;i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  }
  free(tmpbuf);
  return 0;
}



int b_bidir_gather(double *sendbuf, double *recvbuf, struct args *args, 
                 int rank, int size) {

  int offset;
  int finished = FALSE;
  int send;
  int sender = FALSE;
  int i;
  MPI_Status status;
  int tag=0;
  double *tmpbuf;
  double *tmpbuf2;
  int curpos;

  //if(!rank) printf("doing bidir4...\n");

  send = 0;
  offset = 1;
  curpos = rank;
  
  tmpbuf = (double*)malloc(args->numelem*size*sizeof(double));
  tmpbuf2 = (double*)malloc(args->numelem*size*sizeof(double));

  for(i=0;i<args->numelem;i++) recvbuf[rank*args->numelem+i] = sendbuf[i];

  finished = FALSE;

  while(!finished) {
    
    for (i=0;i<size;i++) {
      if (!(i % offset)) {
        send ^= 1;
      }
      if ((rank == i) && send) {
        sender = TRUE;
      }
    }

    for (i=0;i<args->numelem*offset;i++) {
      tmpbuf2[i] = recvbuf[curpos*args->numelem+i];
    }

    if(sender) {
      
      MPI_Send(tmpbuf2, args->numelem*offset, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD);
      MPI_Recv(tmpbuf, args->numelem*offset, MPI_DOUBLE, rank+offset, tag, 
               MPI_COMM_WORLD, &status);

      for (i=0;i<args->numelem*offset;i++) {
        recvbuf[(curpos+offset)*(args->numelem)+i] = tmpbuf[i];
      }
      
    } else {
      MPI_Recv(tmpbuf, args->numelem*offset, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD, &status);
      MPI_Send(tmpbuf2, args->numelem*offset, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD);

      for (i=0;i<args->numelem*offset;i++) {
        recvbuf[(curpos-offset)*(args->numelem)+i] = tmpbuf[i];
      }
      curpos = curpos-offset;
    }
    
    tag++;

    offset *= 2;
    if(offset >= size) {
      finished = TRUE;
    } else {
      sender = FALSE;
    }
  } 

  if((rank==0) && (args->output)){
    for (i=0;i<args->numelem*size;i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  }
  free(tmpbuf);
  free(tmpbuf2);
  return 0;


}


int b_rec_gather(double *sendbuf, double *recvbuf, struct args *args, 
               int rank, int size) {
 
  int *startelem;
  int done = FALSE;
  int offset;
  int send = 0;
  int sender = FALSE;
  int leftover;
  int *blocksize;
  int tmpblock;
  int i,j,k;
  int sendsize;
  double *tmpbuf;
  double *tmpbuf2;
  int tag = 0;
  int activepos;
  int notactivepos;
  int curpos;
  int numsend;
  int numrecv;
  int count;
  MPI_Status status;


  curpos = rank;
 
  leftover = args->numelem % size;
  tmpblock = args->numelem / size;

  blocksize = (int*)malloc(size*sizeof(int));
  startelem = (int*)malloc(size*sizeof(int));
  tmpbuf = (double*)malloc(args->numelem*size*sizeof(double));
  tmpbuf2 = (double*)malloc(args->numelem*size*sizeof(double));

  for (i=0;i<size;i++) {
    blocksize[i] = tmpblock;
    if (leftover > 0) {
      blocksize[i]++;
      leftover--;
    }
  }
  
  startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + blocksize[i-1];
  }

  for(i=0;i<args->numelem;i++) {
    recvbuf[rank*args->numelem+i] = sendbuf[i];
  }

  activepos = 0;
  sendsize = size/2;
  send = 0;
  offset = 1;

  while (!done) {
    for (i=0;i<size;i++) {
      if (!(i % offset)) {
        send ^= 1;
      }
      if ((rank == i) && send) {
        sender = TRUE;
      }
    }

    if (sender) {
      notactivepos = activepos+sendsize;
    } else {
      notactivepos = activepos;
      activepos = activepos + sendsize;
    }

    numsend = 0;
    numrecv = 0;
    count = 0;

    for (j=0;j<offset;j++) {
      for (i=0;i<sendsize;i++) {
        numsend = numsend + blocksize[activepos+i];
        for (k=0;k<blocksize[activepos+i];k++) {
          tmpbuf[count] = 
            recvbuf[(curpos+j)*args->numelem+startelem[activepos+i]+k];
          count++;
        }
      }
    }

    for (j=0;j<offset;j++) {
      for (i=0;i<sendsize;i++) {
        numrecv = numrecv + blocksize[notactivepos+i];
      }
    }
 
    if (sender) {
      MPI_Send(tmpbuf, numsend, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD);
      MPI_Recv(tmpbuf2, numrecv, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &status);
      count = 0;
      for(j=0;j<offset;j++) {
        for(i=0;i<sendsize;i++) {
          for(k=0;k<blocksize[notactivepos+i];k++) {
            recvbuf[(curpos+offset+j)*args->numelem
                   +startelem[notactivepos+i]
                   +k] = tmpbuf2[count];
            count++;
          }
        }
      }
                   
    } else {      
      MPI_Recv(tmpbuf2, numrecv, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD, &status);
      MPI_Send(tmpbuf, numsend, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD);
      count = 0;
      for(j=0;j<offset;j++) {
        for(i=0;i<sendsize;i++) {
          for(k=0;k<blocksize[notactivepos+i];k++) {
            recvbuf[(curpos-offset+j)*args->numelem
                   +startelem[notactivepos+i]
                   +k] = tmpbuf2[count];
            count++;
          }
        }
      }
      
      curpos = curpos - offset;
    }
    
    if (sendsize == 1) {
      done = TRUE;
    } else {
      activepos = notactivepos;
      sendsize = sendsize / 2;
      offset = offset * 2;
      sender = 0;
      send = 0;    
    }
  }


  done = FALSE;
  sender = FALSE;
  send = 0;

  activepos = notactivepos;

  for (i=0;i<size;i++) {
    if (!(i % offset)) {
      send ^= 1;
    }
    if ((rank == i) && send) {
      sender = TRUE;
    }
  }
  
  if (sender) {
    notactivepos = activepos - sendsize;
  } else {
    notactivepos = activepos + sendsize;
  }

  while(!done) {
    count = 0;
    numsend = 0;
    for (j=0;j<size;j++) {
      for (i=0;i<sendsize;i++) {
        numsend = numsend + blocksize[activepos+i];
        for (k=0;k<blocksize[activepos+i];k++) {
          tmpbuf[count] =
            recvbuf[j*args->numelem
                   +startelem[activepos+i]
                   +k];
          count++;
        }
      }
    }

    numrecv = 0;
    for (j=0;j<size;j++) {
      for (i=0;i<sendsize;i++) {
        numrecv = numrecv + blocksize[notactivepos+i];
      }
    }

    if (sender) {
      MPI_Send(tmpbuf, numsend, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD);
      MPI_Recv(tmpbuf2, numrecv, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &status);

      count=0;
      for (j=0;j<size;j++) {
        for (i=0;i<sendsize;i++) {
          for (k=0;k<blocksize[notactivepos+i];k++) {
            tmpbuf[count] =
              recvbuf[j*args->numelem
                     +startelem[notactivepos+i]
                     +k] = tmpbuf2[count];
            count++;
          }
        }
      }
    } else {
      MPI_Recv(tmpbuf2, numrecv, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD, &status);
      MPI_Send(tmpbuf, numsend, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD);     

      count = 0;
      for (j=0;j<size;j++) {
        for (i=0;i<sendsize;i++) {

          for (k=0;k<blocksize[notactivepos+i];k++) {
            recvbuf[j*args->numelem
                   +startelem[notactivepos+i]
                   +k] = tmpbuf2[count];
            count++;
          }
        }
      }

    }
    
    if (sendsize == size/2) {
      done = TRUE;
    } else {
      if (sender) activepos = notactivepos;
        
      sendsize = sendsize * 2;
      offset = offset / 2;
      sender = 0;
      send = 0;
    

      for (i=0;i<size;i++) {
        if (!(i % offset)) {
          send ^= 1;
        }
        if ((rank == i) && send) {
          sender = TRUE;
        }
      }
      
      if (sender) {
        notactivepos = activepos - sendsize;
      } else {
        notactivepos = activepos + sendsize;
      }
    }
  }


  if((rank==args->testrank) && (args->output)){
    for (j=0;j<size;j++) {
      for (i=0;i<args->numelem;i++) {
        printf(" %f ", recvbuf[j*args->numelem+i]);
      }
      printf("\n");
    }
  }

  free(startelem);
  free(blocksize);
  free(tmpbuf);

  return 0;
}



int b_buck_gather(double *sendbuf, double *recvbuf, struct args *args, 
                int rank, int size) {
  int * blocksize;
  int * startelem;
  double * tmpbuf;
  double * tmpbuf2;
  int i, j , k;
  MPI_Status status;
  int leftover;
  int sendsize;
  int curblock;
  int currecvblock;
  int tmpblock;
  int tag=0;

  blocksize = (int*)malloc(size*sizeof(int));
  startelem = (int*)malloc(size*sizeof(int));
  tmpbuf = (double*)malloc(args->numelem*size*sizeof(double));
  tmpbuf2 = (double*)malloc(args->numelem*size*sizeof(double));

  leftover = args->numelem % size;
  tmpblock = args->numelem / size;

  for (i=0;i<size;i++) {
    blocksize[i] = tmpblock;
    if (leftover > 0) {
      blocksize[i]++;
      leftover--;
    }
  }
  
  startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + blocksize[i-1];
  }

  for(i=0;i<args->numelem;i++) {
    recvbuf[rank*args->numelem+i] = sendbuf[i];
  }

  sendsize = 1;
  curblock = rank;
  currecvblock = (size + (rank-1)) % size;
  
  for(i=0;i<((size-1)*2);i++) {
    for(k=0;k<sendsize;k++) {
      for(j=0;j<blocksize[curblock];j++) 
        tmpbuf2[k*blocksize[curblock]+j] = 
          recvbuf[((size+(rank-k))%size)*args->numelem
                 +startelem[curblock]+j];    
    }
    
    if (!rank) {
      MPI_Recv(tmpbuf, sendsize*blocksize[currecvblock], MPI_DOUBLE, size-1, 
                tag, 
                MPI_COMM_WORLD, &status);
      MPI_Send(tmpbuf2, sendsize*blocksize[curblock], MPI_DOUBLE, 
              (rank+1)%size, 
              tag, MPI_COMM_WORLD);
    } else {
      MPI_Send(tmpbuf2, sendsize*blocksize[curblock], MPI_DOUBLE, 
              (rank+1)%size, 
              tag, MPI_COMM_WORLD);
      MPI_Recv(tmpbuf, sendsize*blocksize[currecvblock], MPI_DOUBLE, rank-1, 
                tag, 
                MPI_COMM_WORLD, &status);
    }
    tag++;

    for(k=0;k<sendsize;k++) {
      for(j=0;j<blocksize[currecvblock];j++) 
        recvbuf[((size+(rank-k-1))%size)*args->numelem
               +startelem[currecvblock]+j] = 
          tmpbuf[k*blocksize[currecvblock]+j];    
    }
      
    
    curblock = currecvblock;
    currecvblock = (size + (currecvblock-1)) % size;
    if (sendsize < (size - 1)) sendsize++;
    
  }

  if((rank==args->testrank) && (args->output)){
    for (j=0;j<size;j++) {
      for (i=0;i<args->numelem;i++) {
        printf(" %f ", recvbuf[j*args->numelem+i]);
      }
      printf("\n");
    }
  }

  return 0;
  
}





int b_pipe_gather(double *sendbuf, double *recvbuf, struct args *args, 
                int rank, int size) {
  double * tmpbuf;
  int i, j;
  MPI_Status status;
  int tag=0;
  
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));


  for(i=0;i<args->numelem;i++) {
    recvbuf[rank*args->numelem+i] = sendbuf[i];
  } 


  for (i=0;i<size;i++) {
    if (rank == i) {
      for(j=0;j<size;j++) {
        if (j != i) {
          MPI_Send(sendbuf, args->numelem, MPI_DOUBLE, j,
                    tag, MPI_COMM_WORLD); 
        }
      }
    } else {
      MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, i, tag,  
                MPI_COMM_WORLD, &status); 
    }
    if (rank != i) {
      for (j=0;j<args->numelem;j++) { 
        recvbuf[i*args->numelem+j] = tmpbuf[j]; 
      }
    } 
  }

  if((rank==args->testrank) && (args->output)){
    for (j=0;j<size;j++) {
      for (i=0;i<args->numelem;i++) {
        printf(" %f ", recvbuf[j*args->numelem+i]);
      }
      printf("\n");
    }
  }      
  return 0;

}












