#include "combine.h"
#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>

#define TRUE 1
#define FALSE 0
#define MAXCHAR 100

// Ring All-Reduce

int ring_reduce(double *sendbuf, double *recvbuf, struct args *args, int rank, 
                int size) {
  int tag = 0;
  int i,j;
  double *tmpbuf;
  MPI_Request request;
  MPI_Status status;
  
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for(j=0;j<args->numelem;j++) {
    recvbuf[j] = sendbuf[j];
    tmpbuf[j] = 0;
  }
  
  // Send data to the next node

  MPI_Isend(sendbuf, args->numelem, MPI_DOUBLE, (rank+1)%size, tag, 
            MPI_COMM_WORLD, &request);
  
  // Receive data from previous node and combine it with local data

  if (!rank) {
    MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, size-1, tag, MPI_COMM_WORLD, 
              &request);
  } else {
    MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, rank-1, tag, MPI_COMM_WORLD, 
              &request);
  }
  MPI_Wait(&request, &status);

  for(j=0;j<args->numelem;j++) {
    recvbuf[j] += tmpbuf[j];
  }
  
  // Keep receiving, combining and passing on

  for(i=2;i<size;i++) {
    MPI_Isend(tmpbuf, args->numelem, MPI_DOUBLE, (rank+1)%size, tag, MPI_COMM_WORLD,
              &request);
    if (!rank) {
      MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, size-1, tag, 
                MPI_COMM_WORLD, &request);      
    } else {
      MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, rank-1, tag, 
                MPI_COMM_WORLD, &request);
    } 
    
    tag = i;
     
    MPI_Wait(&request, &status);
    
    for(j=0;j<args->numelem;j++) {
      recvbuf[j] += tmpbuf[j];
    }
  } 
  
  // print results if -o argument defined
  
  if((rank==args->testrank) && (args->output)){
    for (i=0;i<args->numelem;i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  }
  return 0;
}


// Fan-in Fan-out All-Reduce

int fan_reduce (double *sendbuf, double *recvbuf, struct args *args, int rank, 
                int size) {
  int i; 
  int tag=0;
  int finished;
  int offset;
  int checkrank;
  double *tmpbuf;
  MPI_Status status;
  MPI_Request request;


  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for (i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  } 
  
  if (size > 1) {
    // initialise tree traversing variables
    finished = FALSE;
    offset = 1;
    checkrank = rank;

    // Fan-in

    while(!finished) {
      if (checkrank%2 == 0){
	if ((rank + offset) < size) {
	  MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, rank+offset, tag, 
		   MPI_COMM_WORLD, &request);
          MPI_Wait(&request, &status);
	  for (i=0;i<args->numelem;i++) {
	    recvbuf[i] += tmpbuf[i];
	  } 
	}
      } else {
	// if my checkrank is odd numbered, I will send my data
	MPI_Isend(recvbuf, args->numelem, MPI_DOUBLE, rank-offset, tag, 
                  MPI_COMM_WORLD, &request);
        MPI_Wait(&request, &status);
	finished = TRUE;
      }
      
      if (!rank) {
	if (offset*2 >= size) {
	  finished = TRUE;
	}
      }
      
      // Number of nodes to send and receive data cut by half
      checkrank = checkrank/2;
      // offset between ranks is doubled
      offset = offset*2;
    }
  }
  
  // Fan-out

  finished = FALSE;

  while (!finished) {
    if (rank%offset == 0) {
      if ((rank+offset/2) < size) {
        //rank%offset : send(rank+offset/2);
        MPI_Isend(recvbuf, args->numelem, MPI_DOUBLE, rank+offset/2, tag, 
                 MPI_COMM_WORLD, &request);
        MPI_Wait(&request, &status);
      }
    }
    if((rank-offset/2)%offset == 0) {
      if ((rank-offset/2) >= 0) {
        //((rank-offset/2)%offset) : recv(rank-offset/2);
        MPI_Irecv(recvbuf, args->numelem, MPI_DOUBLE, rank-offset/2, tag, 
                 MPI_COMM_WORLD, &request);
        MPI_Wait(&request, &status);
      }
    }
    //if (!rank) printf("gothere\n");
    if (offset == 1) {
      finished = TRUE;
    } else {
      offset = offset/2;
    }
    
  }


  if((rank==args->testrank) && (args->output)){
    for (i=0;i<args->numelem;i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  }

  return 0;
}    
  
// Bi-Directional Exchange All-Reduce

int bidir_reduce (double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size) {

  int offset;
  int finished = FALSE;
  int send;
  int sender = FALSE;
  int i;
  double * tmpbuf;
  MPI_Status status;
  MPI_Request request;
  int tag=0;

  send = 0;
  offset = 1;
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for (i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }

  // finished is finished when all data is exchanged

  while(!finished) {
    
    // define sender pairs
    for (i=0;i<size;i++) {
      if (!(i % offset)) {
        send ^= 1;
      }
      if ((rank == i) && send) {
        sender = TRUE;
      }
    }
    
    // do data exchange
    if(sender) {
      MPI_Isend(recvbuf, args->numelem, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &request);
      MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &request);
      
      
    } else {
      MPI_Isend(recvbuf, args->numelem, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD, &request);
      MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD, &request);
    }
    
    MPI_Wait(&request, &status);
    tag++;

    // combine received data
    for(i=0;i<args->numelem;i++) {
      recvbuf[i] += tmpbuf[i];
    }   

    offset *= 2;
    if(offset >= size) {
      finished = TRUE;
    } else {
      sender = FALSE;
    }
  } 

  // print results if -i argument defined

  if((rank==args->testrank) && (args->output)){
    for (i=0;i<args->numelem;i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  }
  
  return 0;


}

// Recursive-Halving, Recursive-Doubling All-Reduce

int rec_reduce (double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size) {
  int *startelem;
  int done = FALSE;
  int offset;
  int send = 0;
  int sender = FALSE;
  int leftover;
  int *blocksize;
  int tmpblock;
  int i;
  int sendsize;
  double *tmpbuf2;
  int tag = 0;
  int activepos;
  int notactivepos;
  int curnumelem;
  int curnumrecelem;
  double *tmpbuf;

  MPI_Status status;
  MPI_Request request;
 
  leftover = args->numelem % size;
  tmpblock = args->numelem / size;

  blocksize = (int*)malloc(size*sizeof(int));
  startelem = (int*)malloc(size*sizeof(int));
  tmpbuf2 = (double*)malloc(args->numelem*sizeof(double));
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  // I first define the block sizes and calculate the start element of 
  // each block

  for(i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }

  for (i=0;i<size;i++) {
    blocksize[i] = tmpblock;
    if (leftover > 0) {
      blocksize[i]++;
      leftover--;
    }
  }
  
  startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + blocksize[i-1];
  }

  // Initialise values

  activepos = 0;
  sendsize = size/2;
  send = 0;
  offset = 1;

  while (!done) {
    // Create sender pairs
    for (i=0;i<size;i++) {
      if (!(i % offset)) {
        send ^= 1;
      }
      if ((rank == i) && send) {
        sender = TRUE;
      }
    }

    // active position defines the position in which the elements are to send
    if (sender) {
      notactivepos = activepos+sendsize;
    } else {
      notactivepos = activepos;
      activepos = activepos + sendsize;
    }

    // calculate the number of elements to send and receive

    curnumelem = 0;

    for (i=activepos; i< activepos + sendsize;i++) {
      curnumelem = curnumelem + blocksize[i];
    }

    for (i=0;i<curnumelem;i++) {
      tmpbuf2[i] = recvbuf[i+startelem[activepos]];
    } 

    curnumrecelem = 0;

    for (i=notactivepos; i< notactivepos + sendsize; i++) {
      curnumrecelem = curnumrecelem + blocksize[i];
    }

    // do data exchange

    if (sender) {
      MPI_Isend(tmpbuf2, curnumelem, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &request);
      MPI_Irecv(tmpbuf, curnumrecelem, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &request);
    } else {
      MPI_Isend(tmpbuf2, curnumelem, MPI_DOUBLE, rank-offset, tag, 
               MPI_COMM_WORLD, &request);
      MPI_Irecv(tmpbuf, curnumrecelem, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD, &request);
    }
    
    MPI_Wait(&request, &status);

    // combine collected data

    for(i=0;i<curnumrecelem;i++) {
      recvbuf[i+startelem[notactivepos]] += tmpbuf[i];
    }

    // continue until each node has a complete combined block.

    if (sendsize == 1) {
      done = TRUE;
    } else {
      activepos = notactivepos;
      sendsize = sendsize / 2;
      offset = offset * 2;
      sender = 0;
      send = 0;    
    }
  }

  // Recursive-Doubling (tree broadcast)

  // Re-initialise values

  done = FALSE;
  sender = FALSE;
  send = 0;

  activepos = notactivepos;

  for (i=0;i<size;i++) {
    if (!(i % offset)) {
      send ^= 1;
    }
    if ((rank == i) && send) {
      sender = TRUE;
    }
  }
  
  if (sender) {
    notactivepos = activepos - sendsize;
  } else {
    notactivepos = activepos + sendsize;
  }

  while(!done) {
    // calculate the number of elements to receive and send
    curnumelem = 0;
    
    for (i=activepos; i< activepos + sendsize;i++) {
      curnumelem = curnumelem + blocksize[i];
    }
    
    for (i=0;i<curnumelem;i++) {
      tmpbuf2[i] = recvbuf[i+startelem[activepos]];
    } 

    curnumrecelem = 0;
    
    for (i=notactivepos; i< notactivepos + sendsize; i++) {
      curnumrecelem = curnumrecelem + blocksize[i];
    }

    // do data exchange

    if (sender) {
      MPI_Isend(tmpbuf2, curnumelem, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &request);
      MPI_Irecv(tmpbuf, curnumrecelem, MPI_DOUBLE, rank+offset, tag, 
               MPI_COMM_WORLD, &request);

    } else {
      MPI_Isend(tmpbuf2, curnumelem, MPI_DOUBLE, rank-offset, tag, 
               MPI_COMM_WORLD, &request);
      MPI_Irecv(tmpbuf, curnumrecelem, MPI_DOUBLE, rank-offset, tag, 
               MPI_COMM_WORLD, &request);

    }
    MPI_Wait(&request, &status);

    // save the received data

    for(i=0;i<curnumrecelem;i++) {
      recvbuf[i+startelem[notactivepos]] = tmpbuf[i];
    }
    
    if (sendsize == size/2) {
      done = TRUE;
    } else {
      if (sender) activepos = notactivepos;
        
      sendsize = sendsize * 2;
      offset = offset / 2;
      sender = 0;
      send = 0;

      for (i=0;i<size;i++) {
        if (!(i % offset)) {
          send ^= 1;
        }
        if ((rank == i) && send) {
          sender = TRUE;
        }
      }
      
      if (sender) {
        notactivepos = activepos - sendsize;
      } else {
        notactivepos = activepos + sendsize;
      }
    }
  }

  // print results if -o argument defined

  if((rank==args->testrank) && (args->output)) {
    for(i=0;i<args->numelem;i++) {
      printf(" %f",recvbuf[i]);
    }
    printf("\n");
  }
  
  free(startelem);
  free(blocksize);
  free(tmpbuf2);
  return 0;
}

// Bucket All-Reduce (obsolete)

int buck_reduce (double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size) {

  int leftover;
  int tmpblock;
  int *blocksize;
  int *startelem;
  double *tmpbuf2;
  int i, j;
  int tag = 0;
  int cursendpos;
  int currecpos;
  int numiter;
  int count=0;
  double *tmpbuf;
  MPI_Request request;
  MPI_Status status;

  leftover = args->numelem % size;
  tmpblock = args->numelem / size;

  blocksize = (int*)malloc(size*sizeof(int));
  startelem = (int*)malloc(size*sizeof(int));
  tmpbuf2 = (double*)malloc((tmpblock+1)*sizeof(double));
  tmpbuf = (double*)malloc((tmpblock+1)*sizeof(double));

  for(i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }

  for (i=0;i<size;i++) {
    blocksize[i] = tmpblock;
    if (leftover > 0) {
      blocksize[i]++;
      leftover--;
    }
  }
  
  startelem[0] = 0;
  
  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + blocksize[i-1];
  }

  cursendpos = rank;
  if ((cursendpos - 1) >= 0) {
    currecpos = cursendpos - 1;
  } else {
    currecpos = size - 1;
  }
  
  numiter = (size - 1)*2;
  
  for(i=0;i<numiter;i++) {

    for(j=0;j<blocksize[cursendpos];j++) 
      tmpbuf2[j] = recvbuf[startelem[cursendpos]+j];    

    MPI_Isend(tmpbuf2, blocksize[cursendpos], MPI_DOUBLE, (rank+1)%size, 
              tag, MPI_COMM_WORLD,&request);
    
    if (!rank) {
      MPI_Irecv(tmpbuf, blocksize[currecpos], MPI_DOUBLE, size-1, tag, 
                MPI_COMM_WORLD, &request);
    } else {
      MPI_Irecv(tmpbuf, blocksize[currecpos], MPI_DOUBLE, rank-1, tag, 
                MPI_COMM_WORLD, &request);
    }
    tag++;

    MPI_Wait(&request, &status);
    
    for(j=0;j<blocksize[currecpos];j++) {
      if (count < size-1) {
        recvbuf[startelem[currecpos]+j] += tmpbuf[j];
      } else {
        recvbuf[startelem[currecpos]+j] = tmpbuf[j];
      }
    }

    if ((cursendpos - 1) >= 0) {
      cursendpos--;
    } else {
      cursendpos = size - 1;
    }

    if ((cursendpos - 1) >= 0) {
      currecpos = cursendpos - 1;
    } else {
      currecpos = size - 1;
    }
    count++;
  }

  if((rank==args->testrank) && (args->output)){
    for(i=0;i<args->numelem;i++) {
      printf(" %f",recvbuf[i]);
    }
    printf("\n");
  }

  free(startelem);
  free(blocksize);
  free(tmpbuf2);  
  return 0;
}

// MPI Implementation All-Reduce

int mpi_reduce(double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size) {
  int i;
  MPI_Allreduce(sendbuf, recvbuf, args->numelem, MPI_DOUBLE, MPI_SUM, 
                MPI_COMM_WORLD);

  if((rank==args->testrank) && (args->output)){
    for(i=0;i<args->numelem;i++) {
      printf(" %f",recvbuf[i]);
    }
    printf("\n");
  }

  return 0;
}

// Pipeline All-Reduce

int pipe_reduce(double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size) {
  int i, j;
  double * tmpbuf;
  int tag=0;
  MPI_Request request;
  MPI_Status status;
  
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for(i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }

  // each node sends the receive buffer value

  for (i=0;i<size-1;i++) {
    if (rank == i) {
      MPI_Isend(recvbuf, args->numelem, MPI_DOUBLE, (i+1)%size, 
            tag, MPI_COMM_WORLD,&request);
      MPI_Wait(&request, &status);
    } else {
      if (rank == ((i+1) % size)) {
        MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, i, tag, 
                MPI_COMM_WORLD, &request);
        MPI_Wait(&request, &status);
	// when a node collects the data combine it with it current 
	// receive buffer
        for (j=0;j<args->numelem;j++) {
          recvbuf[j] += tmpbuf[j];
        }
      }
    }
  }

  for (i=0;i<size-1;i++) {
    if (rank == i) {
      MPI_Irecv(recvbuf, args->numelem, MPI_DOUBLE, (size-1+i) % size, tag, 
                MPI_COMM_WORLD, &request);
      MPI_Wait(&request, &status);
    } else {
      if (rank == ((size-1+i) % size)) {
        MPI_Isend(recvbuf, args->numelem, MPI_DOUBLE, i, 
                  tag, MPI_COMM_WORLD,&request);
        MPI_Wait(&request, &status);
      }
    }
  }
  
  // print results if -o argument defined

  if((rank==args->testrank) && (args->output)){
    for(i=0;i<args->numelem;i++) {
      printf(" %f",recvbuf[i]);
    }
    printf("\n");
  }

  return 0;
}

// Full Fan-out

int naive_reduce (double *sendbuf, double *recvbuf, struct args *args, int rank, int size) {
  int i, j;
  int tag;
  double * tmpbuf;
  MPI_Status status;
  MPI_Request request;

  // if (!rank) printf("Doing naive reduce\n");

  tag = 0;

  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for(i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }

  // All nodes send and receive data
 
  for(i=1;i<size;i++) {
    // send to (rank + i) mod size
    MPI_Isend(sendbuf, args->numelem, MPI_DOUBLE, (rank + i)%size,
	      tag, MPI_COMM_WORLD, &request);
  }

  for(i=1;i<size;i++) {
    // receive from (rank + i) mod size
    MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, (rank + i)%size,
	      tag, MPI_COMM_WORLD, &request);
    MPI_Wait(&request, &status);
    for (j=0;j<args->numelem;j++) {
      recvbuf[j] += tmpbuf[j];
    }
  }

  // print results if -o argument defined

  if((rank==args->testrank) && (args->output)){
    for(i=0;i<args->numelem;i++) {
      printf(" %f",recvbuf[i]);
    }
    printf("\n");
  }     

  return 0;
}


// ! NEED WORK HERE !
int fourway_reduce (double *sendbuf, double *recvbuf, struct args *args, 
		 int rank, int size) {

  if (!rank) printf("4-way reduce\n");
  return 0;
}

// BLOCKING ALGORITHMS (OBSOLETE)

int b_ring_reduce(double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size) {
  int tag = 0;
  int i,j;
  double *tmpbuf;
  MPI_Status status;
  
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for(j=0;j<args->numelem;j++) {
    recvbuf[j] = sendbuf[j];
    tmpbuf[j] = 0;
  }
  
  if (!rank) {
    MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, size-1, tag, MPI_COMM_WORLD, 
             &status);
    MPI_Send(sendbuf, args->numelem, MPI_DOUBLE, (rank+1)%size, tag, 
            MPI_COMM_WORLD);
  } else {
    MPI_Send(sendbuf, args->numelem, MPI_DOUBLE, (rank+1)%size, tag, 
             MPI_COMM_WORLD);
    MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, rank-1, tag, MPI_COMM_WORLD, 
              &status);
  }

  for(j=0;j<args->numelem;j++) {
    recvbuf[j] += tmpbuf[j];
  }
  
  for(i=2;i<size;i++) {
    if (!rank) {
      MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, size-1, tag, 
               MPI_COMM_WORLD, &status);

      MPI_Send(tmpbuf, args->numelem, MPI_DOUBLE, (rank+1)%size, tag, 
               MPI_COMM_WORLD);
    } else {
      MPI_Send(tmpbuf, args->numelem, MPI_DOUBLE, (rank+1)%size, tag, 
               MPI_COMM_WORLD);
      MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, rank-1, tag, 
                MPI_COMM_WORLD, &status);
    } 
    
    tag = i;
    
    for(j=0;j<args->numelem;j++) {
      recvbuf[j] += tmpbuf[j];
    }
  } 
  
  
  if((rank==args->testrank) && (args->output)){
    for (i=0;i<args->numelem;i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  }
  return 0;
}


int b_fan_reduce (double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size) {
  int i; 
  int tag=0;
  int finished;
  int offset;
  int checkrank;
  double *tmpbuf;
  MPI_Status status;


  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for (i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  } 
  
  if (size > 1) {
    // initialise tree traversing variables
    finished = FALSE;
    offset = 1;
    checkrank = rank;

    // Fan-in

    while(!finished) {
      if (checkrank%2 == 0){
	if ((rank + offset) < size) {
	  MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, rank+offset, tag, 
		   MPI_COMM_WORLD, &status);
	  for (i=0;i<args->numelem;i++) {
	    recvbuf[i] += tmpbuf[i];
	  } 
	}
      } else {
	// if my checkrank is odd numbered, I will send my data
	MPI_Send(recvbuf, args->numelem, MPI_DOUBLE, rank-offset, tag, 
                  MPI_COMM_WORLD);
	finished = TRUE;
      }
      
      if (!rank) {
	if (offset*2 >= size) {
	  finished = TRUE;
	}
      }
      
      // Number of nodes to send and receive data cut by half
      checkrank = checkrank/2;
      // offset between ranks is doubled
      offset = offset*2;
    }
  }
  
  finished = FALSE;

  while (!finished) {
    if (rank%offset == 0) {
      if ((rank+offset/2) < size) {
        //rank%offset : send(rank+offset/2);
        MPI_Send(recvbuf, args->numelem, MPI_DOUBLE, rank+offset/2, tag, 
                 MPI_COMM_WORLD);
      }
    }
    if((rank-offset/2)%offset == 0) {
      if ((rank-offset/2) >= 0) {
        //((rank-offset/2)%offset) : recv(rank-offset/2);
        MPI_Recv(recvbuf, args->numelem, MPI_DOUBLE, rank-offset/2, tag, 
                 MPI_COMM_WORLD, &status);
      }
    }
    //if (!rank) printf("gothere\n");
    if (offset == 1) {
      finished = TRUE;
    } else {
      offset = offset/2;
    }
    
  }


  if((rank==args->testrank) && (args->output)){
    for (i=0;i<args->numelem;i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  }

  return 0;
}    
  

int b_bidir_reduce (double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size) {

  int offset;
  int finished = FALSE;
  int send;
  int sender = FALSE;
  int i;
  double * tmpbuf;
  MPI_Status status;
  int tag=0;

  send = 0;
  offset = 1;
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for (i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }

  while(!finished) {
    
    for (i=0;i<size;i++) {
      if (!(i % offset)) {
        send ^= 1;
      }
      if ((rank == i) && send) {
        sender = TRUE;
      }
    }
    
    if(sender) {
      MPI_Send(recvbuf, args->numelem, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD);
      MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &status);
      
      
    } else {      
      MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD, &status);
      MPI_Send(recvbuf, args->numelem, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD);
    }
    
    tag++;

    for(i=0;i<args->numelem;i++) {
      recvbuf[i] += tmpbuf[i];
    }   

    offset *= 2;
    if(offset >= size) {
      finished = TRUE;
    } else {
      sender = FALSE;
    }
  } 

  if((rank==args->testrank) && (args->output)){
    for (i=0;i<args->numelem;i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  }
  
  return 0;


}

int b_rec_reduce (double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size) {
  int *startelem;
  int done = FALSE;
  int offset;
  int send = 0;
  int sender = FALSE;
  int leftover;
  int *blocksize;
  int tmpblock;
  int i;
  int sendsize;
  double *tmpbuf2;
  int tag = 0;
  int activepos;
  int notactivepos;
  int curnumelem;
  int curnumrecelem;
  double *tmpbuf;

  MPI_Status status;
 
  leftover = args->numelem % size;
  tmpblock = args->numelem / size;

  blocksize = (int*)malloc(size*sizeof(int));
  startelem = (int*)malloc(size*sizeof(int));
  tmpbuf2 = (double*)malloc(args->numelem*sizeof(double));
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for(i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }

  for (i=0;i<size;i++) {
    blocksize[i] = tmpblock;
    if (leftover > 0) {
      blocksize[i]++;
      leftover--;
    }
  }
  
  startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + blocksize[i-1];
  }

  activepos = 0;
  sendsize = size/2;
  send = 0;
  offset = 1;

  while (!done) {
    for (i=0;i<size;i++) {
      if (!(i % offset)) {
        send ^= 1;
      }
      if ((rank == i) && send) {
        sender = TRUE;
      }
    }

    if (sender) {
      notactivepos = activepos+sendsize;
    } else {
      notactivepos = activepos;
      activepos = activepos + sendsize;
    }

    curnumelem = 0;

    for (i=activepos; i< activepos + sendsize;i++) {
      curnumelem = curnumelem + blocksize[i];
    }

    for (i=0;i<curnumelem;i++) {
      tmpbuf2[i] = recvbuf[i+startelem[activepos]];
    } 

    curnumrecelem = 0;

    for (i=notactivepos; i< notactivepos + sendsize; i++) {
      curnumrecelem = curnumrecelem + blocksize[i];
    }

    if (sender) {
      MPI_Send(tmpbuf2, curnumelem, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD);
      MPI_Recv(tmpbuf, curnumrecelem, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &status);
    } else {
      MPI_Recv(tmpbuf, curnumrecelem, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD, &status);
      MPI_Send(tmpbuf2, curnumelem, MPI_DOUBLE, rank-offset, tag, 
               MPI_COMM_WORLD);
    }

    for(i=0;i<curnumrecelem;i++) {
      recvbuf[i+startelem[notactivepos]] += tmpbuf[i];
    }

    if (sendsize == 1) {
      done = TRUE;
    } else {
      activepos = notactivepos;
      sendsize = sendsize / 2;
      offset = offset * 2;
      sender = 0;
      send = 0;    
    }
  }


  done = FALSE;
  sender = FALSE;
  send = 0;

  activepos = notactivepos;

  for (i=0;i<size;i++) {
    if (!(i % offset)) {
      send ^= 1;
    }
    if ((rank == i) && send) {
      sender = TRUE;
    }
  }
  
  if (sender) {
    notactivepos = activepos - sendsize;
  } else {
    notactivepos = activepos + sendsize;
  }

  while(!done) {

    curnumelem = 0;
    
    for (i=activepos; i< activepos + sendsize;i++) {
      curnumelem = curnumelem + blocksize[i];
    }
    
    for (i=0;i<curnumelem;i++) {
      tmpbuf2[i] = recvbuf[i+startelem[activepos]];
    } 

    curnumrecelem = 0;
    
    for (i=notactivepos; i< notactivepos + sendsize; i++) {
      curnumrecelem = curnumrecelem + blocksize[i];
    }


    if (sender) {
      MPI_Send(tmpbuf2, curnumelem, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD);
      MPI_Recv(tmpbuf, curnumrecelem, MPI_DOUBLE, rank+offset, tag, 
               MPI_COMM_WORLD, &status);

    } else {
      MPI_Recv(tmpbuf, curnumrecelem, MPI_DOUBLE, rank-offset, tag, 
               MPI_COMM_WORLD, &status);
      MPI_Send(tmpbuf2, curnumelem, MPI_DOUBLE, rank-offset, tag, 
               MPI_COMM_WORLD);
    }

    for(i=0;i<curnumrecelem;i++) {
      recvbuf[i+startelem[notactivepos]] = tmpbuf[i];
    }
    
    if (sendsize == size/2) {
      done = TRUE;
    } else {
      if (sender) activepos = notactivepos;
        
      sendsize = sendsize * 2;
      offset = offset / 2;
      sender = 0;
      send = 0;

      for (i=0;i<size;i++) {
        if (!(i % offset)) {
          send ^= 1;
        }
        if ((rank == i) && send) {
          sender = TRUE;
        }
      }
      
      if (sender) {
        notactivepos = activepos - sendsize;
      } else {
        notactivepos = activepos + sendsize;
      }
    }
  }

  if((rank==args->testrank) && (args->output)) {
    for(i=0;i<args->numelem;i++) {
      printf(" %f",recvbuf[i]);
    }
    printf("\n");
  }
  
  free(startelem);
  free(blocksize);
  free(tmpbuf2);
  return 0;
}

int b_buck_reduce (double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size) {

  int leftover;
  int tmpblock;
  int *blocksize;
  int *startelem;
  double *tmpbuf2;
  int i, j;
  int tag = 0;
  int cursendpos;
  int currecpos;
  int numiter;
  int count=0;
  double *tmpbuf;
  MPI_Status status;

  leftover = args->numelem % size;
  tmpblock = args->numelem / size;

  blocksize = (int*)malloc(size*sizeof(int));
  startelem = (int*)malloc(size*sizeof(int));
  tmpbuf2 = (double*)malloc((tmpblock+1)*sizeof(double));
  tmpbuf = (double*)malloc((tmpblock+1)*sizeof(double));

  for(i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }

  for (i=0;i<size;i++) {
    blocksize[i] = tmpblock;
    if (leftover > 0) {
      blocksize[i]++;
      leftover--;
    }
  }
  
  startelem[0] = 0;
  
  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + blocksize[i-1];
  }

  cursendpos = rank;
  if ((cursendpos - 1) >= 0) {
    currecpos = cursendpos - 1;
  } else {
    currecpos = size - 1;
  }
  
  numiter = (size - 1)*2;
  
  for(i=0;i<numiter;i++) {

    for(j=0;j<blocksize[cursendpos];j++) 
      tmpbuf2[j] = recvbuf[startelem[cursendpos]+j];    


    
    if (!rank) {
      MPI_Recv(tmpbuf, blocksize[currecpos], MPI_DOUBLE, size-1, tag, 
                MPI_COMM_WORLD, &status);
      MPI_Send(tmpbuf2, blocksize[cursendpos], MPI_DOUBLE, (rank+1)%size, 
               tag, MPI_COMM_WORLD);
    } else {
      MPI_Send(tmpbuf2, blocksize[cursendpos], MPI_DOUBLE, (rank+1)%size, 
               tag, MPI_COMM_WORLD);      
      MPI_Recv(tmpbuf, blocksize[currecpos], MPI_DOUBLE, rank-1, tag, 
                MPI_COMM_WORLD, &status);
    }
    
    for(j=0;j<blocksize[currecpos];j++) {
      if (count < size-1) {
        recvbuf[startelem[currecpos]+j] += tmpbuf[j];
      } else {
        recvbuf[startelem[currecpos]+j] = tmpbuf[j];
      }
    }

    if ((cursendpos - 1) >= 0) {
      cursendpos--;
    } else {
      cursendpos = size - 1;
    }

    if ((cursendpos - 1) >= 0) {
      currecpos = cursendpos - 1;
    } else {
      currecpos = size - 1;
    }
    count++;
  }

  if((rank==args->testrank) && (args->output)){
    for(i=0;i<args->numelem;i++) {
      printf(" %f",recvbuf[i]);
    }
    printf("\n");
  }

  free(startelem);
  free(blocksize);
  free(tmpbuf2);  
  return 0;
}

int b_pipe_reduce(double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size) {
  int i, j;
  double * tmpbuf;
  int tag=0;
  MPI_Status status;
  
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for(i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }

  for (i=0;i<size-1;i++) {
    if (rank == i) {
      MPI_Send(recvbuf, args->numelem, MPI_DOUBLE, (i+1)%size, 
            tag, MPI_COMM_WORLD);
    } else {
      if (rank == ((i+1) % size)) {
        MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, i, tag, 
                MPI_COMM_WORLD, &status);
        for (j=0;j<args->numelem;j++) {
          recvbuf[j] += tmpbuf[j];
        }
      }
    }
  }

  for (i=0;i<size-1;i++) {
    if (rank == i) {
      MPI_Recv(recvbuf, args->numelem, MPI_DOUBLE, (size-1+i) % size, tag, 
                MPI_COMM_WORLD, &status);
    } else {
      if (rank == ((size-1+i) % size)) {
        MPI_Send(recvbuf, args->numelem, MPI_DOUBLE, i, 
                  tag, MPI_COMM_WORLD);
      }
    }
  }
  
  if((rank==args->testrank) && (args->output)){
    for(i=0;i<args->numelem;i++) {
      printf(" %f",recvbuf[i]);
    }
    printf("\n");
  }

  return 0;
}














